import sys
import math
from lark import Lark
from lark.visitors import Interpreter

l = Lark(r"""
int: /\d+/
char: "'" /./
quote: "[" code "]"
string: /"[^"]*"/
var: /[a-z]/
!simple: ";" | ":" | "$" | "%" | "@" | "\\" | "O" | "+" | "-" | "*" | "/" | "_" | "&" | "|" | "~" | "=" | ">" | "!" | "?" | "#" | "^" | "," | "." | "B"
_inst: int | char | string | quote | simple | var
code: _inst*
%ignore /\s+|{[^}]*}/ 
""", parser="lalr", start="code")

variables = {}
stack = []

class Factor(Interpreter):
    def int(self, tree):
        stack.append(int(tree.children[0]))

    def char(self, tree):
        stack.append(ord(tree.children[0]))

    def quote(self, tree):
        stack.append(lambda: self.visit(tree.children[0]))

    def string(self, tree):
        print(tree.children[0][1:-1], end="")

    def var(self, tree):
        stack.append(tree.children[0].value)

    def simple(self, tree):
        match tree.children[0].value:
            case "$":
                stack.append(stack[-1])
            case "%":
                stack.pop()
            case "\\":
                stack.append(stack.pop(-2))
            case "@":
                stack.append(stack.pop(-3))
            case "O":
                stack.append(stack[-stack.pop()-1])
            case "+":
                stack.append(stack.pop() + stack.pop())
            case "-":
                stack.append(stack.pop(-2) - stack.pop())
            case "*":
                stack.append(stack.pop() * stack.pop())
            case "/":
                stack.append(math.trunc(stack.pop(-2) / stack.pop()))
            case "_":
                stack[-1] *= -1
            case "&":
                stack.append(stack.pop() & stack.pop())
            case "|":
                stack.append(stack.pop() | stack.pop())
            case "~":
                stack.append(~stack.pop())
            case ">":
                stack.append(-1 if stack.pop() < stack.pop() else 0)
            case "=":
                stack.append(-1 if stack.pop() == stack.pop() else 0)
            case "!":
                stack.pop()()
            case "?":
                f = stack.pop()
                if stack.pop():
                    f()
            case "#":
                b = stack.pop()
                c = stack.pop()
                while (c(), stack.pop())[1]:
                    b()
            case ":":
                variables[stack.pop()] = stack.pop(-2)
            case ";":
                stack.append(variables[stack.pop()])
            case "^":
                stack.append(sys.stdin.buffer.read(1)[0] or -1)
            case ",":
                sys.stdout.buffer.write(bytes([stack.pop() % 256]))
            case ".":
                print(stack.pop(), end="")
            case "B":
                sys.stdout.flush()

    def code(self, tree):
        for x in tree.children:
            self.visit(x)

Factor().visit(l.parse(open(sys.argv[1]).read()))
